"""Migration entity."""
from abc import ABC, abstractmethod
from dataclasses import dataclass
from enum import Enum, auto
from typing import Any, Dict, List, Set

from butterfree.clients import AbstractClient
from butterfree.configs.logger import __logger
from butterfree.load.writers.writer import Writer
from butterfree.transform import FeatureSet

logger = __logger("database_migrate", True)


@dataclass
class Diff:
    """DataClass to help identifying different types of diff between schemas."""

    class Kind(Enum):
        """Mapping actions to take given a difference between columns of a schema."""

        ADD = auto()
        ALTER_KEY = auto()
        ALTER_TYPE = auto()
        DROP = auto()

    column: str
    kind: Kind
    value: Any

    def __hash__(self) -> int:
        return hash((self.column, self.kind, self.value))

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, type(self)):
            raise NotImplementedError
        return (
            self.column == other.column
            and self.kind == other.kind
            and self.value == other.value
        )


class DatabaseMigration(ABC):
    """Abstract base class for Migrations."""

    def __init__(self, client: AbstractClient) -> None:
        self._client = client

    @abstractmethod
    def _get_create_table_query(
        self, columns: List[Dict[str, Any]], table_name: str
    ) -> Any:
        """Creates desired statement to create a table.

        Args:
            columns: object that contains column's schemas.
            table_name: table name.

        Returns:
            Create table query.

        """
        pass

    @abstractmethod
    def _get_alter_table_add_query(self, columns: List[Diff], table_name: str) -> str:
        """Creates desired statement to add columns to a table.

        Args:
            columns: list of Diff objects with ADD kind.
            table_name: table name.

        Returns:
            Alter table query.

        """
        pass

    @abstractmethod
    def _get_alter_table_drop_query(self, columns: List[Diff], table_name: str) -> str:
        """Creates desired statement to drop columns from a table.

        Args:
            columns: list of Diff objects with DROP kind.
            table_name: table name.

        Returns:
            Drop columns from a given table query.

        """
        pass

    @abstractmethod
    def _get_alter_column_type_query(self, column: Diff, table_name: str) -> str:
        """Creates desired statement to alter columns' types.

        Args:
            columns: list of Diff objects with ALTER_TYPE kind.
            table_name: table name.

        Returns:
            Alter column type query.

        """
        pass

    def _get_queries(
        self, schema_diff: Set[Diff], table_name: str, write_on_entity: bool = None
    ) -> Any:
        """Create the desired queries for migration.

        Args:
            schema_diff: list of Diff objects.
            table_name: table name.

        Returns:
            List of queries.

        """
        add_items = []
        drop_items = []
        alter_type_items = []
        alter_key_items = []

        for diff in schema_diff:
            if diff.kind == Diff.Kind.ADD:
                add_items.append(diff)
            elif diff.kind == Diff.Kind.ALTER_TYPE:
                alter_type_items.append(diff)
            elif diff.kind == Diff.Kind.DROP:
                drop_items.append(diff)
            elif diff.kind == Diff.Kind.ALTER_KEY:
                alter_key_items.append(diff)

        queries = []
        if add_items:
            alter_table_add_query = self._get_alter_table_add_query(
                add_items, table_name
            )
            queries.append(alter_table_add_query)
        if drop_items:
            if not write_on_entity:
                drop_columns_query = self._get_alter_table_drop_query(
                    drop_items, table_name
                )
                queries.append(drop_columns_query)
        if alter_type_items:
            for item in alter_type_items:
                alter_column_types_query = self._get_alter_column_type_query(
                    item, table_name
                )
                queries.append(alter_column_types_query)
        if alter_key_items:
            logger.warning(
                "The 'change the primary key column' action is not supported by Spark."
            )

        return queries

    def create_query(
        self,
        fs_schema: List[Dict[str, Any]],
        table_name: str,
        db_schema: List[Dict[str, Any]] = None,
        write_on_entity: bool = None,
    ) -> Any:
        """Create a query regarding a data source.

        Returns:
            The desired queries for the given database.

        """
        if not db_schema:
            return [self._get_create_table_query(fs_schema, table_name)]

        schema_diff = self._get_diff(fs_schema, db_schema)

        return self._get_queries(schema_diff, table_name, write_on_entity)

    @staticmethod
    def _get_diff(
        fs_schema: List[Dict[str, Any]], db_schema: List[Dict[str, Any]],
    ) -> Set[Diff]:
        """Gets schema difference between feature set and the table of a given db.

        Args:
            fs_schema: object that contains feature set's schemas.
            db_schema: object that contains the table of a given db schema.

        Returns:
            Object with schema differences.

        """
        db_columns = set(item.get("column_name") for item in db_schema)
        fs_columns = set(item.get("column_name") for item in fs_schema)

        add_columns = fs_columns - db_columns
        drop_columns = db_columns - fs_columns

        # This could be way easier to write (and to read) if the schemas were a simple
        # Dict[str, Any] where each key would be the column name itself...
        # but changing that could break things so:
        # TODO version 2 change get schema to return a dict(columns, properties)
        add_type_columns = dict()
        alter_type_columns = dict()
        alter_key_columns = dict()
        for fs_item in fs_schema:
            if fs_item.get("column_name") in add_columns:
                add_type_columns.update(
                    {fs_item.get("column_name"): fs_item.get("type")}
                )
            for db_item in db_schema:
                if fs_item.get("column_name") == db_item.get("column_name"):
                    if fs_item.get("type") != db_item.get("type"):
                        if fs_item.get("primary_key") is True:
                            logger.warning(
                                "Type changes are not applied to "
                                "columns that are the primary key."
                            )
                        alter_type_columns.update(
                            {fs_item.get("column_name"): fs_item.get("type")}
                        )
                    if fs_item.get("primary_key") != db_item.get("primary_key"):
                        alter_key_columns.update(
                            {fs_item.get("column_name"): fs_item.get("primary_key")}
                        )
                    break

        schema_diff = set(
            Diff(str(col), kind=Diff.Kind.ADD, value=value)
            for col, value in add_type_columns.items()
        )
        schema_diff |= set(
            Diff(str(col), kind=Diff.Kind.DROP, value=None) for col in drop_columns
        )
        schema_diff |= set(
            Diff(str(col), kind=Diff.Kind.ALTER_TYPE, value=value)
            for col, value in alter_type_columns.items()
        )
        schema_diff |= set(
            Diff(str(col), kind=Diff.Kind.ALTER_KEY, value=None)
            for col, value in alter_key_columns.items()
        )
        return schema_diff

    def _get_schema(
        self, table_name: str, database: str = None
    ) -> List[Dict[str, Any]]:
        """Get a table schema in the respective database.

        Args:
            table_name: Table name to get schema.

        Returns:
            Schema object.
        """
        try:
            db_schema = self._client.get_schema(table_name, database)
        except Exception:  # noqa
            db_schema = []
        return db_schema

    def apply_migration(
        self, feature_set: FeatureSet, writer: Writer, debug_mode: bool
    ) -> None:
        """Apply the migration in the respective database.

        Args:
            feature_set: the feature set.
            writer: the writer being used to load the feature set.
            debug_mode: if active, it brings up the queries generated.
        """
        logger.info(f"Migrating feature set: {feature_set.name}")

        table_name = (
            feature_set.name if not writer.write_to_entity else feature_set.entity
        )

        fs_schema = writer.db_config.translate(feature_set.get_schema())
        db_schema = self._get_schema(table_name, writer.database)

        queries = self.create_query(
            fs_schema, table_name, db_schema, writer.write_to_entity
        )

        if debug_mode:
            print(
                "#### DEBUG MODE ###\n"
                f"Feature set: {feature_set.name}\n"
                "Queries:\n"
                f"{queries}"
            )
        else:
            for q in queries:
                logger.info(f"Applying this query: {q} ...")
                self._client.sql(q)

            logger.info(f"Feature Set migration finished successfully.")

            # inform in drone console which feature set was migrated
            print(f"The {feature_set.name} feature set was migrated.")



"""Writer entity."""

from abc import ABC, abstractmethod
from functools import reduce
from typing import Any, Callable, Dict, List

from pyspark.sql.dataframe import DataFrame

from butterfree.clients import SparkClient
from butterfree.configs.db import AbstractWriteConfig
from butterfree.hooks import HookableComponent
from butterfree.transform import FeatureSet


class Writer(ABC, HookableComponent):
    """Abstract base class for Writers.

    Args:
        spark_client: client for spark connections with external services.

    """

    def __init__(
        self,
        db_config: AbstractWriteConfig,
        debug_mode: bool = False,
        interval_mode: bool = False,
        write_to_entity: bool = False,
    ) -> None:
        super().__init__()
        self.db_config = db_config
        self.transformations: List[Dict[str, Any]] = []
        self.debug_mode = debug_mode
        self.interval_mode = interval_mode
        self.write_to_entity = write_to_entity

    def with_(
        self, transformer: Callable[..., DataFrame], *args: Any, **kwargs: Any
    ) -> "Writer":
        """Define a new transformation for the Writer.

        All the transformations are used when the method consume is called.

        Args:
            transformer: method that receives a dataframe and output a
                dataframe.
            *args: args for the transformer.
            **kwargs: kwargs for the transformer.

        Returns:
            Reader object with new transformation

        """
        new_transformation = {
            "transformer": transformer,
            "args": args if args else (),
            "kwargs": kwargs if kwargs else {},
        }
        self.transformations.append(new_transformation)
        return self

    def _apply_transformations(self, df: DataFrame) -> DataFrame:
        return reduce(
            lambda result_df, transformation: transformation["transformer"](
                result_df, *transformation["args"], **transformation["kwargs"]
            ),
            self.transformations,
            df,
        )

    @abstractmethod
    def write(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient,
    ) -> Any:
        """Loads the data from a feature set into the Feature Store.

        Feature Store could be Online or Historical.

        Args:
            feature_set: object processed with feature set metadata.
            dataframe: Spark dataframe containing data from a feature set.
            spark_client: client for Spark connections with external services.

        """

    @abstractmethod
    def check_schema(
        self, client: Any, dataframe: DataFrame, table_name: str, database: str = None
    ) -> DataFrame:
        """Instantiate the schema check hook to check schema between dataframe and database.

        Args:
            client: client for Spark or Cassandra connections with external services.
            dataframe: Spark dataframe containing data from a feature set.
            table_name: table name where the dataframe will be saved.
            database: database name where the dataframe will be saved.
        """

    @abstractmethod
    def validate(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient
    ) -> Any:
        """Calculate dataframe rows to validate data into Feature Store.

        Args:
            feature_set: object processed with feature set metadata.
            dataframe: Spark dataframe containing data from a feature set.
            spark_client: client for Spark connections with external services.

        Raises:
            AssertionError: if validation fails.

        """



"""Aggregated Transform entity."""
from collections import namedtuple
from typing import List, Tuple

from pyspark.sql import DataFrame
from pyspark.sql.functions import col, expr, when

from butterfree.transform.transformations.transform_component import TransformComponent
from butterfree.transform.utils.function import Function


class AggregatedTransform(TransformComponent):
    """Specifies an aggregation.

    This transformation needs to be used within an AggregatedFeatureSet. Unlike the
    other transformations, this class won't have a transform method implemented.

    The idea behing aggregating is that, in spark, we should execute all aggregation
    functions after a single groupby. So an AggregateFeatureSet will have many Features
    with AggregatedTransform. If each one of them needs to apply a groupby.agg(), then
    we must join all the results in the end, making this computation extremely slow.

    Now, the AggregateFeatureSet will collect all Features' AggregatedTransform
    definitions and run, at once, a groupby.agg(*aggregations).

    This class helps defining on a feature, which aggregation function will be applied
    to build a new aggregated column. Allowed aggregations are registered under the
     allowed_aggregations property.

    Attributes:
        functions: namedtuple with aggregation function and data type.
        filter_expression: sql boolean expression to be used inside agg function.
            The filter expression can be used to aggregate some column only with
            records that obey certain condition. Has the same behaviour of the
            following SQL expression: `agg(case when filter_expression then col end)`

    Example:
        >>> from butterfree.transform.transformations import AggregatedTransform
        >>> from butterfree.transform.features import Feature
        >>> from butterfree.constants import DataType
        >>> from butterfree.transform.utils import Function
        >>> import pyspark.sql.functions as F
        >>> feature = Feature(
        ...     name="feature",
        ...     description="aggregated transform",
        ...     transformation=AggregatedTransform(
        ...         functions=[
        ...                    Function(F.avg, DataType.DOUBLE),
        ...                    Function(F.stddev_pop, DataType.DOUBLE)],
        ...     ),
        ...     from_column="somenumber",
        ...)
        >>> feature.get_output_columns()
        ['feature__avg', 'feature__stddev_pop']
        >>> feature.transform(anydf)
        NotImplementedError: ...
    """

    def __init__(self, functions: List[Function], filter_expression: str = None):
        super(AggregatedTransform, self).__init__()
        self.functions = functions
        self.filter_expression = filter_expression

    @property
    def aggregations(self) -> List[Tuple]:
        """Aggregated spark columns."""
        column_name = self._parent.from_column or self._parent.name

        # if transform has a filter expression apply inside agg function
        # if not, use just the target column name inside agg function
        expression = (
            when(expr(self.filter_expression), col(column_name))
            if self.filter_expression
            else column_name
        )
        Function = namedtuple("Function", ["function", "data_type"])

        return [
            Function(f.func(expression), f.data_type.spark,) for f in self.functions
        ]

    def _get_output_name(self, function: object) -> str:
        if not hasattr(function, "__name__"):
            feature_name = self._parent.name
            raise AttributeError(
                f"""Anonymous functions are not supported on AggregatedTransform.
                    Check feature {feature_name} transforms.
                """
            )

        base_name = "__".join([self._parent.name, str(function.__name__).lower()])
        return base_name

    @property
    def output_columns(self) -> List[str]:
        """Columns names generated by the transformation."""
        return [self._get_output_name(f.func) for f in self.functions]

    def transform(self, dataframe: DataFrame) -> DataFrame:
        """(NotImplemented) Performs a transformation to the feature pipeline.

        For the AggregatedTransform, the transformation won't be applied without
        using an AggregatedFeatureSet.

        Args:
            dataframe: input dataframe.

        Raises:
            NotImplementedError.
        """
        raise NotImplementedError(
            "AggregatedTransform won't be used outside an AggregatedFeatureSet, "
            "meaning the responsibility of aggregating and apply the transformation is "
            "now over the FeatureSet component. This should optimize the ETL process."
        )



"""Holds the Online Feature Store writer class."""

import os
from typing import Any, Dict, List, Union

from pyspark.sql import DataFrame, Window
from pyspark.sql.functions import col, row_number
from pyspark.sql.streaming import StreamingQuery

from butterfree.clients import SparkClient
from butterfree.configs.db import AbstractWriteConfig, CassandraConfig
from butterfree.constants.columns import TIMESTAMP_COLUMN
from butterfree.hooks import Hook
from butterfree.hooks.schema_compatibility import CassandraTableSchemaCompatibilityHook
from butterfree.load.writers.writer import Writer
from butterfree.transform import FeatureSet


class OnlineFeatureStoreWriter(Writer):
    """Enable writing feature sets into the Online Feature Store.

    Attributes:
        db_config: Spark configuration for connect databases.
            For more information check the module 'butterfree.db.configs'.
        debug_mode: "dry run" mode, write the result to a temporary view.
        write_to_entity: option to write the data to the entity table.
            With this option set to True, the writer will write the feature set to
            a table with the name equal to the entity name, defined on the pipeline.
            So, it WILL NOT write to a table with the name of the feature set, as it
            normally does.

    Example:
        Simple example regarding OnlineFeatureStoreWriter class instantiation.
        We can instantiate this class without db configurations, so the class get the
        CassandraConfig() where it provides default configurations about CassandraDB.

    >>> spark_client = SparkClient()
    >>> writer = OnlineFeatureStoreWriter()
    >>> writer.write(feature_set=feature_set,
       ...           dataframe=dataframe,
       ...           spark_client=spark_client)

        However, we can define the db configurations and provide them to
        OnlineFeatureStoreWriter.

    >>> spark_client = SparkClient()
    >>> config = CassandraConfig(mode="overwrite",
        ...                      format_="parquet",
        ...                      keyspace="keyspace_name")

    >>> writer = OnlineFeatureStoreWriter(db_config=config)
    >>> writer.write(feature_set=feature_set,
       ...           dataframe=dataframe,
       ...           spark_client=spark_client)

        For what settings you can use on CassandraConfig and default settings,
        to read CassandraConfig class.

        We can instantiate OnlineFeatureStoreWriter class to validate the writers,
        using the default or custom configs.

    >>> spark_client = SparkClient()
    >>> writer = OnlineFeatureStoreWriter()
    >>> writer.validate(feature_set=feature_set,
       ...              dataframe=dataframe,
       ...              spark_client=spark_client)

        Both methods (writer and validate) will need the Spark Client,
        Feature Set and DataFrame, to write or to validate,
        according to OnlineFeatureStoreWriter class arguments.

        There's an important aspect to be highlighted here: if you're using
        the incremental mode, we do not check if your data is the newest before
        writing to the online feature store.

        This behavior is known and will be fixed soon.
    """

    __name__ = "Online Feature Store Writer"

    def __init__(
        self,
        db_config: AbstractWriteConfig = None,
        database: str = None,
        debug_mode: bool = False,
        write_to_entity: bool = False,
        interval_mode: bool = False,
        check_schema_hook: Hook = None,
    ):
        super(OnlineFeatureStoreWriter, self).__init__(
            db_config or CassandraConfig(), debug_mode, interval_mode, write_to_entity
        )
        self.check_schema_hook = check_schema_hook
        self.database = database

    @staticmethod
    def filter_latest(dataframe: DataFrame, id_columns: List[Any]) -> DataFrame:
        """Filters latest data from the dataframe.

        Args:
            dataframe: spark dataframe containing data from a feature set.
            id_columns: unique identifier column set for this feature set.

        Returns:
            dataframe: contains only latest data for each unique id in the
                feature set.

        """
        if TIMESTAMP_COLUMN not in dataframe.columns:
            raise KeyError("DataFrame must have a 'ts' column to order by.")
        if id_columns is None or not id_columns:
            raise ValueError("Users must provide the unique identifiers.")
        missing = [c for c in id_columns if c not in dataframe.columns]
        if missing:
            raise KeyError(f"{missing} not found in the DataFrame.")

        window = Window.partitionBy(*id_columns).orderBy(col(TIMESTAMP_COLUMN).desc())
        return (
            dataframe.select(col("*"), row_number().over(window).alias("rn"),)
            .filter(col("rn") == 1)
            .drop("rn")
        )

    def _write_stream(
        self,
        feature_set: FeatureSet,
        dataframe: DataFrame,
        spark_client: SparkClient,
        table_name: str,
    ) -> StreamingQuery:
        """Writes the dataframe in streaming mode."""
        checkpoint_folder = (
            f"{feature_set.name}__on_entity" if self.write_to_entity else table_name
        )
        checkpoint_path = (
            os.path.join(
                self.db_config.stream_checkpoint_path,
                feature_set.entity,
                checkpoint_folder,
            )
            if self.db_config.stream_checkpoint_path
            else None
        )
        streaming_handler = spark_client.write_stream(
            dataframe,
            processing_time=self.db_config.stream_processing_time,
            output_mode=self.db_config.stream_output_mode,
            checkpoint_path=checkpoint_path,
            format_=self.db_config.format_,
            mode=self.db_config.mode,
            **self.db_config.get_options(table_name),
        )
        return streaming_handler

    @staticmethod
    def _write_in_debug_mode(
        table_name: str, dataframe: DataFrame, spark_client: SparkClient
    ) -> Union[StreamingQuery, None]:
        """Creates a temporary table instead of writing to the real feature store."""
        return spark_client.create_temporary_view(
            dataframe=dataframe, name=f"online_feature_store__{table_name}"
        )

    def write(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient,
    ) -> Union[StreamingQuery, None]:
        """Loads the latest data from a feature set into the Feature Store.

        Args:
            feature_set: object processed with feature set metadata.
            dataframe: Spark dataframe containing data from a feature set.
            spark_client: client for Spark connections with external services.

        Returns:
            Streaming handler if writing streaming df, None otherwise.

        If the debug_mode is set to True, a temporary table with a name in the format:
        `online_feature_store__my_feature_set` will be created instead of writing to
        the real online feature store. If dataframe is streaming this temporary table
        will be updated in real time.

        """
        table_name = feature_set.entity if self.write_to_entity else feature_set.name

        if dataframe.isStreaming:
            dataframe = self._apply_transformations(dataframe)
            if self.debug_mode:
                return self._write_in_debug_mode(
                    table_name=table_name,
                    dataframe=dataframe,
                    spark_client=spark_client,
                )
            return self._write_stream(
                feature_set=feature_set,
                dataframe=dataframe,
                spark_client=spark_client,
                table_name=table_name,
            )

        latest_df = self.filter_latest(
            dataframe=dataframe, id_columns=feature_set.keys_columns
        )

        latest_df = self._apply_transformations(latest_df)

        if self.debug_mode:
            return self._write_in_debug_mode(
                table_name=table_name, dataframe=latest_df, spark_client=spark_client
            )

        return spark_client.write_dataframe(
            dataframe=latest_df,
            format_=self.db_config.format_,
            mode=self.db_config.mode,
            **self.db_config.get_options(table_name),
        )

    def validate(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient
    ) -> None:
        """Calculate dataframe rows to validate data into Feature Store.

        Args:
            feature_set: object processed with feature set metadata.
            dataframe: Spark dataframe containing data from a feature set.
            spark_client: client for Spark connections with external services.

        Raises:
            AssertionError: if validation fails.

        """
        # validation with online feature store can be done right now, since
        # this database can be updated by an streaming job while our ETL is running
        # therefore, creating in consistencies between this ETL feature set count and
        # the data already written to the FS.
        # TODO how to run data validations when a database has concurrent writes.
        pass

    def get_db_schema(self, feature_set: FeatureSet) -> List[Dict[Any, Any]]:
        """Get desired database schema.

        Args:
            feature_set: object processed with feature set metadata.

        Returns:
            Desired database schema.

        """
        db_schema = self.db_config.translate(feature_set.get_schema())
        return db_schema

    def check_schema(
        self, client: Any, dataframe: DataFrame, table_name: str, database: str = None
    ) -> DataFrame:
        """Instantiate the schema check hook to check schema between dataframe and database.

        Args:
            client: client for Spark or Cassandra connections with external services.
            dataframe: Spark dataframe containing data from a feature set.
            table_name: table name where the dataframe will be saved.
            database: database name where the dataframe will be saved.
        """
        if not self.check_schema_hook:
            self.check_schema_hook = CassandraTableSchemaCompatibilityHook(
                client, table_name
            )

        return self.check_schema_hook.run(dataframe)



"""FeatureSet entity."""
import itertools
from functools import reduce
from typing import Any, Dict, List, Optional

import pyspark.sql.functions as F
from pyspark.sql import Window
from pyspark.sql.dataframe import DataFrame

from butterfree.clients import SparkClient
from butterfree.constants.columns import TIMESTAMP_COLUMN
from butterfree.dataframe_service import IncrementalStrategy
from butterfree.hooks import HookableComponent
from butterfree.transform.features import Feature, KeyFeature, TimestampFeature
from butterfree.transform.transformations import (
    AggregatedTransform,
    SparkFunctionTransform,
)


class FeatureSet(HookableComponent):
    """Holds metadata about the feature set and constructs the final dataframe.

    Attributes:
        name:  name of the feature set.
        entity: business context tag for the feature set, an entity for which we
            are creating all these features.
        description: details about the feature set purpose.
        keys: key features to define this feature set.
            Values for keys (may be a composition) should be unique on each
            moment in time (controlled by the TimestampFeature).
        timestamp: A single feature that define a timestamp for each observation
            in this feature set.
        features: features to compose the feature set.

    Example:
        This an example regarding the feature set definition. All features
        and its transformations are defined.

    >>> from butterfree.transform import FeatureSet
    >>> from butterfree.transform.features import (
    ...     Feature,
    ...     KeyFeature,
    ...     TimestampFeature,
    ...)
    >>> from butterfree.transform.transformations import (
    ...     SparkFunctionTransform,
    ...     CustomTransform,
    ... )
    >>> from butterfree.constants import DataType
    >>> from butterfree.transform.utils import Function
    >>> import pyspark.sql.functions as F

    >>> def divide(df, fs, column1, column2):
    ...     name = fs.get_output_columns()[0]
    ...     df = df.withColumn(name, F.col(column1) / F.col(column2))
    ...     return df

    >>> feature_set = FeatureSet(
    ...    name="feature_set",
    ...    entity="entity",
    ...    description="description",
    ...    features=[
    ...        Feature(
    ...            name="feature1",
    ...            description="test",
    ...            transformation=SparkFunctionTransform(
    ...                 functions=[
    ...                            Function(F.avg, DataType.DOUBLE),
    ...                            Function(F.stddev_pop, DataType.DOUBLE)]
    ...             ).with_window(
    ...                 partition_by="id",
    ...                 order_by=TIMESTAMP_COLUMN,
    ...                 mode="fixed_windows",
    ...                 window_definition=["2 minutes", "15 minutes"],
    ...             ),
    ...        ),
    ...        Feature(
    ...            name="divided_feature",
    ...            description="unit test",
    ...            transformation=CustomTransform(
    ...                transformer=divide, column1="feature1", column2="feature2",
    ...            ),
    ...        ),
    ...    ],
    ...    keys=[KeyFeature(name="id", description="The user's Main ID or device ID")],
    ...    timestamp=TimestampFeature(),
    ...)

    >>> feature_set.construct(dataframe=dataframe)

    This last method (construct) will execute the feature set, computing all the
    defined transformations.

    There's also a functionality regarding the construct method within the scope
    of FeatureSet called filter_duplicated_rows. We drop rows that have repeated
    values over key columns and timestamp column, we do this in order to reduce
    our dataframe (regarding the number of rows). A detailed explation of this
    method can be found at filter_duplicated_rows docstring.
    """

    def __init__(
        self,
        name: str,
        entity: str,
        description: str,
        keys: List[KeyFeature],
        timestamp: TimestampFeature,
        features: List[Feature],
    ) -> None:
        super().__init__()
        self.name = name
        self.entity = entity
        self.description = description
        self.keys = keys
        self.timestamp = timestamp
        self.features = features
        self.incremental_strategy = IncrementalStrategy(column=TIMESTAMP_COLUMN)

    @property
    def name(self) -> str:
        """Name of the feature set."""
        return self.__name

    @name.setter
    def name(self, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError("name must be a string with the feature set label.")
        self.__name = value

    @property
    def entity(self) -> str:
        """Business context tag for the feature set."""
        return self.__entity

    @entity.setter
    def entity(self, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError(
                "entity must be a string tagging the feature set business context."
            )
        self.__entity = value

    @property
    def description(self) -> str:
        """Details about the feature set purpose."""
        return self.__description

    @description.setter
    def description(self, value: str) -> None:
        if not isinstance(value, str):
            raise ValueError(
                "description must be a string with the feature set details."
            )
        self.__description = value

    @staticmethod
    def _get_features_columns(*features: Feature) -> List[str]:
        return list(itertools.chain(*[k.get_output_columns() for k in features]))

    @property
    def keys(self) -> List[KeyFeature]:
        """Key features to define this feature set."""
        return self.__keys

    @keys.setter
    def keys(self, value: List[KeyFeature]) -> None:
        if not isinstance(value, list) or not all(
            isinstance(item, KeyFeature) for item in value
        ):
            raise ValueError("keys needs to be a list of KeyFeature objects.")

        key_columns = self._get_features_columns(*value)
        if len(key_columns) != len(set(key_columns)):
            raise KeyError("key columns will have duplicates.")

        self.__keys = value

    @property
    def timestamp(self) -> TimestampFeature:
        """Defines a timestamp for each observation in this feature set."""
        return self.__timestamp

    @timestamp.setter
    def timestamp(self, value: TimestampFeature) -> None:
        if not isinstance(value, TimestampFeature):
            raise ValueError("timestamp needs to be a TimestampFeature object.")

        timestamp_columns = value.get_output_columns()
        if len(timestamp_columns) > 1:
            raise ValueError("TimestampFeature will produce multiple output columns.")

        self.__timestamp = value

    @property
    def features(self) -> List[Feature]:
        """Features to compose the feature set."""
        return self.__features

    @features.setter
    def features(self, value: List[Feature]) -> None:
        if not isinstance(value, list) or not all(
            isinstance(item, Feature) for item in value
        ):
            raise ValueError("features needs to be a list of Feature objects.")

        feature_columns = self._get_features_columns(*value)
        if len(feature_columns) != len(set(feature_columns)):
            raise KeyError("feature columns will have duplicates.")

        if self._has_aggregated_transform(value):
            raise ValueError(
                "You can't define a feature within aggregated transform. "
                "You need to use Aggregated Feature Set."
            )

        self.__features = value

    @property
    def keys_columns(self) -> List[str]:
        """Name of the columns of all keys in feature set."""
        return self._get_features_columns(*self.keys)

    @property
    def timestamp_column(self) -> str:
        """Name of the timestamp column in feature set."""
        return self._get_features_columns(self.timestamp).pop()

    @property
    def features_columns(self) -> List[str]:
        """Name of the columns of all features in feature set."""
        return self._get_features_columns(*self.features)

    @property
    def columns(self) -> List[str]:
        """All data columns within this feature set.

        This references all data columns that will be created by the construct
        method, given keys, timestamp and features of this feature set.

        Returns:
            List of column names built in this feature set.

        """
        return self.keys_columns + [self.timestamp_column] + self.features_columns

    def get_schema(self) -> List[Dict[str, Any]]:
        """Get feature set schema.

        Returns:
            List of dicts regarding cassandra feature set schema.

        """
        schema = []

        for f in self.keys + [self.timestamp]:
            for c in self._get_features_columns(f):
                schema.append(
                    {
                        "column_name": c,
                        "type": f.dtype.spark,
                        "primary_key": True if isinstance(f, KeyFeature) else False,
                    }
                )

        for f in self.features:  # type: ignore
            name = self._get_features_columns(f)
            if isinstance(f.transformation, SparkFunctionTransform):
                type = [
                    fc.data_type.spark
                    for fc in f.transformation.functions
                    for _ in range(len(f.transformation._windows or [None]))
                ]
            else:
                type = [f.dtype.spark]

            for n, dt in zip(name, type):
                schema.append({"column_name": n, "type": dt, "primary_key": False})

        return schema

    @staticmethod
    def _has_aggregated_transform(features: List[Feature]) -> bool:
        """Aggregated Transform mode check.

        Checks if there's a rolling window mode within the scope of the
        AggregatedTransform.

        Returns:
            True if there's a rolling window aggregation mode.

        """
        return any(
            [
                isinstance(feature.transformation, AggregatedTransform)
                for feature in features
            ]
        )

    def _filter_duplicated_rows(self, df: DataFrame) -> DataFrame:
        """Filter dataframe duplicated rows.

        Attributes:
            df: transformed dataframe.

        Returns:
            Spark dataframe with filtered rows.

        Example:
            Suppose, for instance, that the transformed dataframe received
            by the construct method has the following rows:

            +---+----------+--------+--------+--------+
            | id| timestamp|feature1|feature2|feature3|
            +---+----------+--------+--------+--------+
            |  1|         1|       0|    null|       1|
            |  1|         2|       0|       1|       1|
            |  1|         3|    null|    null|    null|
            |  1|         4|       0|       1|       1|
            |  1|         5|       0|       1|       1|
            |  1|         6|    null|    null|    null|
            |  1|         7|    null|    null|    null|
            +---+-------------------+--------+--------+

            We will then create four columns, the first one, rn_by_key_columns
            (rn1) is the row number over a key columns partition ordered by timestamp.
            The second, rn_by_all_columns (rn2), is the row number over all columns
            partition (also ordered by timestamp). The third column,
            lag_rn_by_key_columns (lag_rn1), returns the last occurrence of the
            rn_by_key_columns over all columns partition. The last column, diff, is
            the difference between rn_by_key_columns and lag_rn_by_key_columns:

            +---+----------+--------+--------+--------+----+----+--------+-----+
            | id| timestamp|feature1|feature2|feature3| rn1| rn2| lag_rn1| diff|
            +---+----------+--------+--------+--------+----+----+--------+-----+
            |  1|         1|       0|    null|       1|   1|   1|    null| null|
            |  1|         2|       0|       1|       1|   2|   1|    null| null|
            |  1|         3|    null|    null|    null|   3|   1|    null| null|
            |  1|         4|       0|       1|       1|   4|   2|       2|    2|
            |  1|         5|       0|       1|       1|   5|   3|       4|    1|
            |  1|         6|    null|    null|    null|   6|   2|       3|    3|
            |  1|         7|    null|    null|    null|   7|   3|       6|    1|
            +---+----------+--------+--------+--------+----+----+--------+-----+

            Finally, this dataframe will then be filtered with the following condition:
            rn_by_all_columns = 1 or diff > 1 and only the original columns will be
            returned:

            +---+----------+--------+--------+--------+
            | id| timestamp|feature1|feature2|feature3|
            +---+----------+--------+--------+--------+
            |  1|         1|       0|    null|       1|
            |  1|         2|       0|       1|       1|
            |  1|         3|    null|    null|    null|
            |  1|         4|       0|       1|       1|
            |  1|         6|    null|    null|    null|
            +---+----------+--------+--------+--------+

        """
        window_key_columns = Window.partitionBy(
            self.keys_columns  # type: ignore
        ).orderBy(TIMESTAMP_COLUMN)
        window_all_columns = Window.partitionBy(
            self.keys_columns + self.features_columns  # type: ignore
        ).orderBy(TIMESTAMP_COLUMN)

        df = (
            df.withColumn("rn_by_key_columns", F.row_number().over(window_key_columns))
            .withColumn("rn_by_all_columns", F.row_number().over(window_all_columns))
            .withColumn(
                "lag_rn_by_key_columns",
                F.lag("rn_by_key_columns", 1).over(window_all_columns),
            )
            .withColumn(
                "diff", F.col("rn_by_key_columns") - F.col("lag_rn_by_key_columns")
            )
        )
        df = df.filter("rn_by_all_columns = 1 or diff > 1")

        return df.select([column for column in self.columns])

    def define_start_date(self, start_date: str = None) -> Optional[str]:
        """Get feature set start date.

        Args:
            start_date: start date regarding source dataframe.

        Returns:
            start date.
        """
        return start_date

    def construct(
        self,
        dataframe: DataFrame,
        client: SparkClient,
        end_date: str = None,
        num_processors: int = None,
        start_date: str = None,
    ) -> DataFrame:
        """Use all the features to build the feature set dataframe.

        After that, there's the caching of the dataframe, however since cache()
        in Spark is lazy, an action is triggered in order to force persistence.

        Args:
            dataframe: input dataframe to be transformed by the features.
            client: client responsible for connecting to Spark session.
            start_date: user defined start date.
            end_date: user defined end date.
            num_processors: cluster total number of processors for repartitioning.

        Returns:
            Spark dataframe with all the feature columns.

        """
        if not isinstance(dataframe, DataFrame):
            raise ValueError("source_df must be a dataframe")

        pre_hook_df = self.run_pre_hooks(dataframe)

        output_df = reduce(
            lambda df, feature: feature.transform(df),
            self.keys + [self.timestamp] + self.features,
            pre_hook_df,
        ).select(*self.columns)

        if not output_df.isStreaming:
            output_df = self._filter_duplicated_rows(output_df)
            output_df.cache().count()

        output_df = self.incremental_strategy.filter_with_incremental_strategy(
            dataframe=output_df, start_date=start_date, end_date=end_date
        )

        post_hook_df = self.run_post_hooks(output_df)

        return post_hook_df



"""Cassandra table schema compatibility Hook definition."""

from pyspark.sql import DataFrame

from butterfree.clients import CassandraClient
from butterfree.constants import DataType
from butterfree.hooks.hook import Hook


class CassandraTableSchemaCompatibilityHook(Hook):
    """Hook to verify the schema compatibility with a Cassandra's table.

    Verifies if all columns presented on the dataframe exists and are the same
    type on the target Cassandra's table.

    Attributes:
        cassandra_client: client to connect to Cassandra DB.
        table: table name.
    """

    def __init__(self, cassandra_client: CassandraClient, table: str):
        self.cassandra_client = cassandra_client
        self.table = table

    def run(self, dataframe: DataFrame) -> DataFrame:
        """Check the schema compatibility from a given Dataframe.

        This method does not change anything on the Dataframe.

        Args:
            dataframe: dataframe to verify schema compatibility.

        Returns:
            unchanged dataframe.

        Raises:
            ValueError if the schemas are incompatible.
        """
        table_schema = self.cassandra_client.get_schema(self.table)
        type_cassandra = [
            type.cassandra
            for field_id in range(len(dataframe.schema.fieldNames()))
            for type in DataType
            if dataframe.schema.fields.__getitem__(field_id).dataType == type.spark
        ]
        schema = [
            {"column_name": f"{column}", "type": f"{type}"}
            for column, type in zip(dataframe.columns, type_cassandra)
        ]

        if not all([column in table_schema for column in schema]):
            raise ValueError(
                "There's a schema incompatibility "
                "between the defined dataframe and the Cassandra table.\n"
                f"Dataframe schema = {schema}"
                f"Target table schema = {table_schema}"
            )
        return dataframe



"""Reader entity."""

from abc import ABC, abstractmethod
from functools import reduce
from typing import Any, Callable, Dict, List, Optional

from pyspark.sql import DataFrame

from butterfree.clients import SparkClient
from butterfree.dataframe_service import IncrementalStrategy
from butterfree.hooks import HookableComponent


class Reader(ABC, HookableComponent):
    """Abstract base class for Readers.

    Attributes:
        id: unique string id for register the reader as a view on the metastore.
        transformations: list os methods that will be applied over the dataframe
            after the raw data is extracted.

    """

    def __init__(self, id: str, incremental_strategy: IncrementalStrategy = None):
        super().__init__()
        self.id = id
        self.transformations: List[Dict[str, Any]] = []
        self.incremental_strategy = incremental_strategy

    def with_(
        self, transformer: Callable[..., DataFrame], *args: Any, **kwargs: Any
    ) -> Any:
        """Define a new transformation for the Reader.

        All the transformations are used when the method consume is called.

        Args:
            transformer: method that receives a dataframe and output a
                dataframe.
            *args: args for the transformer.
            **kwargs: kwargs for the transformer.

        Returns:
            Reader object with new transformation

        """
        new_transformation = {
            "transformer": transformer,
            "args": args if args else (),
            "kwargs": kwargs if kwargs else {},
        }
        self.transformations.append(new_transformation)
        return self

    def with_incremental_strategy(
        self, incremental_strategy: IncrementalStrategy
    ) -> "Reader":
        """Define the incremental strategy for the Reader.

        Args:
            incremental_strategy: definition of the incremental strategy.

        Returns:
            Reader with defined incremental strategy.
        """
        self.incremental_strategy = incremental_strategy
        return self

    @abstractmethod
    def consume(self, client: SparkClient) -> DataFrame:
        """Extract data from target origin.

        Args:
            client: client responsible for connecting to Spark session.

        Returns:
            Dataframe with all the data.

        :return: Spark dataframe
        """

    def build(
        self,
        client: SparkClient,
        columns: List[Any] = None,
        start_date: str = None,
        end_date: str = None,
    ) -> None:
        """Register the data got from the reader in the Spark metastore.

        Create a temporary view in Spark metastore referencing the data
        extracted from the target origin after the application of all the
        defined pre-processing transformations.

        The arguments start_date and end_date are going to be use only when there
        is a defined `IncrementalStrategy` on the `Reader`.

        Args:
            client: client responsible for connecting to Spark session.
            columns: list of tuples for selecting/renaming columns on the df.
            start_date: lower bound to use in the filter expression.
            end_date: upper bound to use in the filter expression.

        """
        column_selection_df = self._select_columns(columns, client)
        transformed_df = self._apply_transformations(column_selection_df)

        if self.incremental_strategy:
            transformed_df = self.incremental_strategy.filter_with_incremental_strategy(
                transformed_df, start_date, end_date
            )

        post_hook_df = self.run_post_hooks(transformed_df)

        post_hook_df.createOrReplaceTempView(self.id)

    def _select_columns(
        self, columns: Optional[List[Any]], client: SparkClient
    ) -> DataFrame:
        df = self.consume(client)
        return df.selectExpr(
            *(
                [
                    f"{old_expression} as {new_column_name}"
                    for old_expression, new_column_name in columns
                ]
                if columns
                else df.columns
            )
        )

    def _apply_transformations(self, df: DataFrame) -> DataFrame:
        return reduce(
            lambda result_df, transformation: transformation["transformer"](
                result_df, *transformation["args"], **transformation["kwargs"]
            ),
            self.transformations,
            df,
        )



"""Migrations' Constants."""
from butterfree.constants import columns

PARTITION_BY = [
    {"column_name": columns.PARTITION_YEAR, "type": "INT"},
    {"column_name": columns.PARTITION_MONTH, "type": "INT"},
    {"column_name": columns.PARTITION_DAY, "type": "INT"},
]



"""Metastore Migration entity."""

from typing import Any, Dict, List

from butterfree.clients import SparkClient
from butterfree.configs import environment
from butterfree.configs.db import MetastoreConfig
from butterfree.constants.migrations import PARTITION_BY
from butterfree.migrations.database_migration.database_migration import (
    DatabaseMigration,
    Diff,
)


class MetastoreMigration(DatabaseMigration):
    """MetastoreMigration class for performing migrations.

    This class implements some methods of the parent DatabaseMigration class and
    has specific methods for query building.
    The MetastoreMigration class will be used, as the name suggests, for applying
    changes to a given Metastore table. There are, however, some remarks that need
    to be highlighted:
        - If an existing feature has its type changed, then it's extremely important to
        make sure that this conversion would not result in data loss;
        - If new features are added to your feature set, then they're going to be added
        to the corresponding Metastore table;
        - Since feature sets can be written both to a feature set and an entity table,
        we're not going to automatically drop features when using entity tables, since
        it means that some features belong to a different feature set. In summary, if
        data is being loaded into an entity table, then users can drop columns manually.
    """

    def __init__(self, database: str = None,) -> None:
        self._db_config = MetastoreConfig()
        self.database = database or environment.get_variable(
            "FEATURE_STORE_HISTORICAL_DATABASE"
        )
        super(MetastoreMigration, self).__init__(SparkClient())

    @staticmethod
    def _get_parsed_columns(columns: List[Diff]) -> List[str]:
        """Parse columns from a list of Diff objects.

        Args:
            columns: list of Diff objects.

        Returns:
            Parsed columns.

        """
        parsed_columns = []
        for col in columns:
            parsed_columns.append(f"{col.column} {col.value}")

        parsed_columns = ", ".join(parsed_columns)  # type: ignore

        return parsed_columns

    def _get_alter_table_add_query(self, columns: List[Diff], table_name: str) -> str:
        """Creates SQL statement to add columns to a table.

        Args:
            columns: list of Diff objects with ADD kind.
            table_name: table name.

        Returns:
            Alter table query.

        """
        parsed_columns = self._get_parsed_columns(columns)

        return (
            f"ALTER TABLE {self.database}.{table_name} "
            f"ADD IF NOT EXISTS columns ({parsed_columns});"
        )

    def _get_alter_column_type_query(self, column: Diff, table_name: str) -> str:
        """Creates SQL statement to alter columns' types.

        Args:
            columns: list of Diff objects with ALTER_TYPE kind.
            table_name: table name.

        Returns:
            Alter column type query.

        """
        parsed_columns = self._get_parsed_columns([column])

        return f"ALTER TABLE {table_name} ALTER COLUMN {parsed_columns};"

    def _get_create_table_query(
        self, columns: List[Dict[str, Any]], table_name: str
    ) -> str:
        """Creates SQL statement to create a table.

        Args:
            columns: object that contains column's schemas.
            table_name: table name.

        Returns:
            Create table query.

        """
        parsed_columns = []
        for col in columns:
            parsed_columns.append(f"{col['column_name']} {col['type']}")
        parsed_columns = ", ".join(parsed_columns)  # type: ignore

        return (
            f"CREATE TABLE IF NOT EXISTS  "
            f"{self.database}.{table_name} ({parsed_columns}) "
            f"PARTITIONED BY ("
            f"{PARTITION_BY[0]['column_name']} {PARTITION_BY[0]['type']}, "
            f"{PARTITION_BY[1]['column_name']} {PARTITION_BY[1]['type']}, "
            f"{PARTITION_BY[2]['column_name']} {PARTITION_BY[2]['type']});"
        )

    def _get_alter_table_drop_query(self, columns: List[Diff], table_name: str) -> str:
        """Creates SQL statement to drop columns from a table.

        Args:
            columns: list of Diff objects with DROP kind.
            table_name: table name.

        Returns:
            Drop columns from a given table query.

        """
        parsed_columns = self._get_parsed_columns(columns)

        return f"ALTER TABLE {table_name} DROP IF EXISTS ({parsed_columns});"



"""SQL Expression Transform entity."""

from typing import List

from pyspark.sql import DataFrame
from pyspark.sql.functions import expr

from butterfree.transform.transformations.transform_component import TransformComponent


class SQLExpressionTransform(TransformComponent):
    """Defines an SQL Expression Transformation.

    Attributes:
        expression: SQL expression defined by the user.

    Example:
        It's necessary to declare the custom SQL query, such that it
        corresponds to operations between existing columns in the dataframe.
        Besides, the usage of SparkSQL functions is also allowed. Finally,
        a "complete select statement", such as "select col_a * col_b from
        my_table", is not necessary, just the simple operations are
        required, for instance "col_a / col_b" or "col_a * col_b".

        >>> from butterfree.transform.features import Feature
        >>> from butterfree.transform.transformations import SQLExpressionTransform
        >>> from butterfree.constants import DataType
        >>> from pyspark import SparkContext
        >>> from pyspark.sql import session
        >>> import pyspark.sql.functions as F
        >>> sc = SparkContext.getOrCreate()
        >>> spark = session.SparkSession(sc)
        >>> df = spark.createDataFrame([(1, "2016-04-11 11:31:11", 200, 200),
        ...                             (1, "2016-04-11 11:44:12", 300, 300),
        ...                             (1, "2016-04-11 11:46:24", 400, 400),
        ...                             (1, "2016-04-11 12:03:21", 500, 500)]
        ...                           ).toDF("id", "timestamp", "feature1", "feature2")
        >>> feature = Feature(
        ...    name="feature",
        ...    description="SQL expression transform usage example",
        ...    dtype=DataType.DOUBLE,
        ...    transformation=SQLExpressionTransform(expression="feature1/feature2"),
        ...)
        >>> feature.transform(df).orderBy("timestamp").show()
        +--------+--------+---+-------------------+----------------------+
        |feature1|feature2| id|          timestamp|feature1_over_feature2|
        +--------+--------+---+-------------------+----------------------+
        |     200|     200|  1|2016-04-11 11:31:11|                   1.0|
        |     300|     300|  1|2016-04-11 11:44:12|                   1.0|
        |     400|     400|  1|2016-04-11 11:46:24|                   1.0|
        |     500|     500|  1|2016-04-11 12:03:21|                   1.0|
        +--------+--------+---+-------------------+----------------------+

    """

    def __init__(
        self, expression: str,
    ):
        super().__init__()
        self.expression = expression

    @property
    def output_columns(self) -> List[str]:
        """Columns generated by the transformation."""
        return [self._parent.name]

    def transform(self, dataframe: DataFrame) -> DataFrame:
        """Performs a transformation to the feature pipeline.

        Args:
            dataframe: input dataframe.

        Returns:
            Transformed dataframe.

        """
        return dataframe.withColumn(self._parent.name, expr(self.expression))



"""Abstract classes for database configurations with spark."""

from abc import ABC, abstractmethod
from typing import Any, Dict, List


class AbstractWriteConfig(ABC):
    """Abstract class for database write configurations with spark."""

    @property
    @abstractmethod
    def database(self) -> str:
        """Database name."""

    @property
    @abstractmethod
    def mode(self) -> Any:
        """Config option "mode" for spark write.

        Args:

        Returns:
          str: mode.

        """

    @property
    @abstractmethod
    def format_(self) -> Any:
        """Config option "format" for spark write.

        Args:

        Returns:
          str: format.

        """

    @abstractmethod
    def translate(self, schema: Any) -> List[Dict[Any, Any]]:
        """Translate feature set spark schema to the corresponding database.

        Args:
          schema: feature set schema

        Returns:
            Corresponding database schema.

        """



"""Holds configurations to read and write with Spark to AWS S3."""

import os
from typing import Any, Dict, List, Optional

from pyspark.sql import DataFrame

from butterfree.configs import environment
from butterfree.configs.db import AbstractWriteConfig
from butterfree.dataframe_service import extract_partition_values


class MetastoreConfig(AbstractWriteConfig):
    """Configuration for Spark metastore database stored.

    By default the configuration is for AWS S3.

    Attributes:
        path: database root location.
        mode: writing mode used be writers.
        format_: expected stored file format.
        file_system: file schema uri, like: s3a, file.

    """

    def __init__(
        self,
        path: str = None,
        mode: str = None,
        format_: str = None,
        file_system: str = None,
    ):
        self.path = path
        self.mode = mode
        self.format_ = format_
        self.file_system = file_system

    @property
    def database(self) -> str:
        """Database name."""
        return "metastore"

    @property
    def path(self) -> Optional[str]:
        """Bucket name."""
        return self.__path

    @path.setter
    def path(self, value: str) -> None:
        self.__path = value or environment.get_variable("FEATURE_STORE_S3_BUCKET")

    @property
    def format_(self) -> Optional[str]:
        """Expected stored file format."""
        return self.__format

    @format_.setter
    def format_(self, value: str) -> None:
        self.__format = value or "parquet"

    @property
    def mode(self) -> Optional[str]:
        """Writing mode used be writers."""
        return self.__mode

    @mode.setter
    def mode(self, value: str) -> None:
        self.__mode = value or "overwrite"

    @property
    def file_system(self) -> Optional[str]:
        """Writing mode used be writers."""
        return self.__file_system

    @file_system.setter
    def file_system(self, value: str) -> None:
        self.__file_system = value or "s3a"

    def get_options(self, key: str) -> Dict[Optional[str], Optional[str]]:
        """Get options for Metastore.

        Options will be a dictionary with the write and read configuration for
        Spark Metastore.

        Args:
            key: path to save data into Metastore.

        Returns:
            Options configuration for Metastore.

        """
        return {
            "mode": self.mode,
            "format_": self.format_,
            "path": os.path.join(f"{self.file_system}://{self.path}/", key),
        }

    def get_path_with_partitions(self, key: str, dataframe: DataFrame) -> List:
        """Get options for AWS S3 from partitioned parquet file.

        Options will be a dictionary with the write and read configuration for
        Spark to AWS S3.

        Args:
            key: path to save data into AWS S3 bucket.
            dataframe: spark dataframe containing data from a feature set.

        Returns:
            A list of string for file-system backed data sources.
        """
        path_list = []
        dataframe_values = extract_partition_values(
            dataframe, partition_columns=["year", "month", "day"]
        )
        for row in dataframe_values:
            path_list.append(
                f"{self.file_system}://{self.path}/{key}/year={row['year']}/"
                f"month={row['month']}/day={row['day']}"
            )

        return path_list

    def translate(self, schema: List[Dict[str, Any]]) -> List[Dict[str, Any]]:
        """Translate feature set spark schema to the corresponding database."""
        spark_sql_mapping = {
            "TimestampType": "TIMESTAMP",
            "BinaryType": "BINARY",
            "BooleanType": "BOOLEAN",
            "DateType": "DATE",
            "DecimalType": "DECIMAL",
            "DoubleType": "DOUBLE",
            "FloatType": "FLOAT",
            "IntegerType": "INT",
            "LongType": "BIGINT",
            "StringType": "STRING",
            "ArrayType(LongType,true)": "ARRAY<BIGINT>",
            "ArrayType(StringType,true)": "ARRAY<STRING>",
            "ArrayType(FloatType,true)": "ARRAY<FLOAT>",
        }
        sql_schema = []
        for features in schema:
            sql_schema.append(
                {
                    "column_name": features["column_name"],
                    "type": spark_sql_mapping[str(features["type"])],
                    "primary_key": features["primary_key"],
                }
            )

        return sql_schema



"""Allowed windows units and lengths in seconds."""

ALLOWED_WINDOWS = {
    "second": 1,
    "seconds": 1,
    "minute": 60,
    "minutes": 60,
    "hour": 3600,
    "hours": 3600,
    "day": 86400,
    "days": 86400,
    "week": 604800,
    "weeks": 604800,
    "year": 29030400,
    "years": 29030400,
}



"""Logger funcion."""

import logging


def __config(json_file_logs: bool = False) -> None:

    if json_file_logs:
        return logging.basicConfig(
            format='{"name": "%(name)s", "timestamp": "%(asctime)-15s", '
            '"level": "%(levelname)s", "message": "%(message)s"}',
            level=logging.INFO,
            filename="../logging.json",
        )
    return logging.basicConfig(
        format="%(name)s:%(asctime)-15s:%(levelname)s:< %(message)s >",
        level=logging.INFO,
    )


def __logger(name: str, file_logs: bool = False) -> logging.Logger:

    __config(file_logs)
    return logging.getLogger(name)



"""Dataframe optimization components regarding Butterfree."""
from butterfree.dataframe_service.incremental_strategy import IncrementalStrategy
from butterfree.dataframe_service.partitioning import extract_partition_values
from butterfree.dataframe_service.repartition import repartition_df, repartition_sort_df

__all__ = [
    "extract_partition_values",
    "IncrementalStrategy",
    "repartition_df",
    "repartition_sort_df",
]



"""SparkClient entity."""

import json
from typing import Any, Dict, List, Optional, Union

from pyspark.sql import DataFrame, DataFrameReader, SparkSession
from pyspark.sql.streaming import DataStreamReader, StreamingQuery
from pyspark.sql.types import StructType

from butterfree.clients import AbstractClient


class SparkClient(AbstractClient):
    """Handle Spark session connection.

    Get query results with SQL, reads and writes data on external systems.

    """

    def __init__(self) -> None:
        self._session: Optional[SparkSession] = None

    @property
    def conn(self) -> SparkSession:
        """Gets or creates an SparkSession.

        Returns:
            Spark session

        """
        if not self._session:
            self._session = SparkSession.builder.getOrCreate()
        return self._session

    def read(
        self,
        format: str,
        path: Optional[Union[str, List[str]]] = None,
        schema: Optional[StructType] = None,
        stream: bool = False,
        **options: Any,
    ) -> DataFrame:
        """Use the SparkSession.read interface to load data into a dataframe.

        Check docs for more information:
            https://spark.apache.org/docs/latest/sql-data-sources-load-save-functions.html#generic-loadsave-functions

        Args:
            format: string with the format to be used by the DataframeReader.
            path: optional string or a list of string for file-system.
            stream:  flag to indicate if data must be read in stream mode.
            schema: an optional pyspark.sql.types.StructType for the input schema.
            options: options to setup the DataframeReader.

        Returns:
            Dataframe

        """
        if not isinstance(format, str):
            raise ValueError("format needs to be a string with the desired read format")
        if path and not isinstance(path, (str, list)):
            raise ValueError("path needs to be a string or a list of string")

        df_reader: Union[
            DataStreamReader, DataFrameReader
        ] = self.conn.readStream if stream else self.conn.read

        df_reader = df_reader.schema(schema) if schema else df_reader

        return df_reader.format(format).load(path=path, **options)  # type: ignore

    def read_table(self, table: str, database: str = None) -> DataFrame:
        """Use the SparkSession.read interface to read a metastore table.

        Args:
            database: name of the metastore database/schema
            table: name of the table in metastore

        Returns:
            Dataframe

        """
        if not isinstance(table, str):
            raise ValueError(
                "table needs to be a string with the name of the registered table"
            )
        return self.conn.read.table(f"{database}.{table}" if database else table)

    def sql(self, query: str) -> DataFrame:
        """Run a query using Spark SQL.

        Args:
            query: Spark SQL query.

        Returns:
            Dataframe

        """
        return self.conn.sql(query)

    @staticmethod
    def write_dataframe(
        dataframe: DataFrame, format_: str, mode: str, **options: Any
    ) -> None:
        """Receive a spark DataFrame and write it.

        Args:
            dataframe: dataframe containing data from a feature set.
            format_: format used to save the dataframe.
            mode: writing modem can be "error", "append", "overwrite" or
                "ignore". For more information:
                [here](https://spark.apache.org/docs/2.3.0/sql-programming-guide.html#save-modes).
            **options: all other options that can be used in a DataFrameWriter.

        """
        if not isinstance(format_, str):
            raise ValueError("format needs to be a string")
        if not isinstance(mode, str):
            raise ValueError("mode needs to be a string")
        dataframe.write.save(format=format_, mode=mode, **options)

    def write_stream(
        self,
        dataframe: DataFrame,
        processing_time: str,
        output_mode: str,
        checkpoint_path: Optional[str],
        format_: str,
        mode: str,
        **options: Any,
    ) -> StreamingQuery:
        """Starts streaming data writing job.

        Args:
            dataframe: Spark dataframe containing data from a feature set.
            processing_time: a processing time interval as a string.
                E.g. '5 seconds', '1 minute'. Set a trigger that runs the
                mini-batch periodically based on the processing time. If the
                effect of processing data as soon as the data arrives, without
                having to wait for the time frame, is desired, the value
                '0 seconds' can be set.
            output_mode: specifies how data of a streaming DataFrame/Dataset is
                written to a streaming sink destination.
            checkpoint_path: path on S3 to save checkpoints for the stream job.
                These checkpoint can be used on the the job re-start to return
                from where it stops.
            format_: format used to save the dataframe.
            mode: writing modem can be "error", "append", "overwrite" or
                "ignore". For more information:
                [here](https://spark.apache.org/docs/2.3.0/sql-programming-guide.html#save-modes).
            **options: all other options that can be used in a DataFrameWriter.

        More information about processing_time, output_mode and checkpoint_path
        can be found in Spark documentation:
        [here](https://spark.apache.org/docs/latest/structured-streaming-programming-guide.html)

        Returns:
            Streaming handler.

        """
        if not dataframe.isStreaming:
            raise ValueError("A stream df is needed to start a streaming job.")

        return (
            dataframe.writeStream.trigger(processingTime=processing_time)
            .outputMode(output_mode)
            .option("checkpointLocation", checkpoint_path)
            .foreachBatch(
                lambda batch_df, _: self.write_dataframe(
                    batch_df, format_, mode, **options
                )
            )
            .start()
        )

    @staticmethod
    def write_table(
        dataframe: DataFrame,
        database: Optional[str],
        table_name: str,
        path: str,
        format_: str = None,
        mode: str = None,
        partition_by: List[str] = None,
        **options: Any,
    ) -> None:
        """Receive a spark DataFrame and write it as a table in metastore.

        Args:
            dataframe: spark dataframe containing data from a feature set.
            database: specified database name.
            table_name: specified table name.
            path: string with the local to save the table.
            format_: string with the format used to save.
            mode: writing mode, it can be: "error", "append", "overwrite" or
                "ignore". More information:
                [here](https://spark.apache.org/docs/2.3.0/sql-programming-guide.html#save-modes).
            partition_by: names of partitioning columns.
            options: all other options that can be used in a DataFrameWriter.

        """
        if not isinstance(database, str):
            raise ValueError("database needs to be a string")
        if not isinstance(table_name, str):
            raise ValueError("table_name needs to be a string")
        if not isinstance(path, str):
            raise ValueError("path needs to be a string of the local to save")

        name = "{}.{}".format(database, table_name)

        dataframe.write.saveAsTable(
            mode=mode,
            format=format_,
            partitionBy=partition_by,
            name=name,
            path=path,
            **options,
        )

    @staticmethod
    def create_temporary_view(dataframe: DataFrame, name: str) -> Any:
        """Create a temporary view from a given dataframe.

        Args:
            dataframe: dataframe to be be queried by the view.
            name: name of the temporary view.

        """
        if not dataframe.isStreaming:
            return dataframe.createOrReplaceTempView(name)
        return dataframe.writeStream.format("memory").queryName(name).start()

    def add_table_partitions(
        self, partitions: List[Dict[str, Any]], table: str, database: str = None
    ) -> None:
        """Add partitions to an existing table.

        Args:
            partitions: partitions to add to the table.
                It's expected a list of partition dicts to add to the table.
                Example: `[{"year": 2020, "month": 8, "day": 14}, ...]`
            table: table to add the partitions.
            database: name of the database where the table is saved.
        """
        for partition_dict in partitions:
            if not all(
                (
                    isinstance(key, str)
                    and (isinstance(value, str) or isinstance(value, int))
                )
                for key, value in partition_dict.items()
            ):
                raise ValueError(
                    "Partition keys must be column names "
                    "and values must be string or int."
                )

        database_expr = f"`{database}`." if database else ""
        key_values_expr = [
            ", ".join(
                [
                    "{} = {}".format(k, v)
                    if not isinstance(v, str)
                    else "{} = '{}'".format(k, v)
                    for k, v in partition.items()
                ]
            )
            for partition in partitions
        ]
        partitions_expr = " ".join(f"PARTITION ( {expr} )" for expr in key_values_expr)
        command = (
            f"ALTER TABLE {database_expr}`{table}` ADD IF NOT EXISTS {partitions_expr}"
        )

        self.conn.sql(command)

    @staticmethod
    def _filter_schema(schema: DataFrame) -> List[str]:
        """Returns filtered schema with the desired information.

        Attributes:
            schema: desired table.

        Returns:
            A list of strings in the format
            ['{"column_name": "example1", type: "Spark_type"}', ...]

        """
        return (
            schema.filter(
                ~schema.col_name.isin(
                    ["# Partition Information", "# col_name", "year", "month", "day"]
                )
            )
            .toJSON()
            .collect()
        )

    def _convert_schema(self, schema: DataFrame) -> List[Dict[str, str]]:
        """Returns schema with the desired information.

        Attributes:
            schema: desired table.

        Returns:
            A list of dictionaries in the format
            [{"column_name": "example1", type: "Spark_type"}, ...]

        """
        schema_list = self._filter_schema(schema)
        converted_schema = []
        for row in schema_list:
            converted_schema.append(json.loads(row))

        return converted_schema

    def get_schema(self, table: str, database: str = None) -> List[Dict[str, str]]:
        """Returns desired table schema.

        Attributes:
            table: desired table.

        Returns:
            A list of dictionaries in the format
            [{"column_name": "example1", type: "Spark_type"}, ...]

        """
        query = f"DESCRIBE {database}.{table} "  # noqa

        response = self.sql(query)

        if not response:
            raise RuntimeError(
                f"No columns found for table: {table}" f"in database: {database}"
            )

        return self._convert_schema(response)



"""Pivot Transform for dataframes."""
from typing import Callable, List, Union

from pyspark.sql import DataFrame, functions
from pyspark.sql.types import DataType

from butterfree.extract.pre_processing import forward_fill


def pivot(
    dataframe: DataFrame,
    group_by_columns: List[str],
    pivot_column: str,
    agg_column: str,
    aggregation: Callable,
    mock_value: Union[float, str] = None,
    mock_type: Union[DataType, str] = None,
    with_forward_fill: bool = False,
) -> DataFrame:
    """Defines a pivot transformation.

    Attributes:
        dataframe: dataframe to be pivoted.
        group_by_columns: list of columns' names to be grouped.
        pivot_column: column to be pivoted.
        agg_column: column to be aggregated by pivoted category.
        aggregation: desired spark aggregation function to be performed.
            An example: spark_agg(col_name). See docs for all spark_agg:
            https://spark.apache.org/docs/2.3.1/api/python/_modules/pyspark/sql/functions.html
        mock_value: value used to make a difference between true nulls resulting from
            the aggregation and empty values from the pivot transformation.
        mock_type: mock_value data type (compatible with spark).
        with_forward_fill: applies a forward fill to null values after the pivot
            operation.

    Example:

        >>> dataframe.orderBy("ts", "id", "amenity").show()
        +---+---+-------+-----+
        | id| ts|amenity|  has|
        +---+---+-------+-----+
        |  1|  1| fridge|false|
        |  1|  1|   oven| true|
        |  1|  1|   pool|false|
        |  2|  2|balcony|false|
        |  1|  3|balcony| null|
        |  1|  4|   oven| null|
        |  1|  4|   pool| true|
        |  1|  5|balcony| true|
        +---+---+-------+-----+

        >>> pivoted = pivot(dataframe, ["id", "ts"], "amenity", "has", functions.first)
        >>> pivoted.orderBy("ts", "id").show()
        +---+---+-------+------+----+-----+
        | id| ts|balcony|fridge|oven| pool|
        +---+---+-------+------+----+-----+
        |  1|  1|   null| false|true|false|
        |  2|  2|  false|  null|null| null|
        |  1|  3|   null|  null|null| null|
        |  1|  4|   null|  null|null| true|
        |  1|  5|   true|  null|null| null|
        +---+---+-------+------+----+-----+

        But, sometimes, you would like to keep the last values that some feature has
        assumed from previous modifications. In this example, amenity "oven" for the
        id=1 was set to null and "pool" was set to true at ts=4. All other amenities
        should then be kept to their actual state at that ts. To do that, we will use
        a technique called forward fill:

        >>> pivoted = pivot(
        ...     dataframe,
        ...     ["id", "ts"],
        ...     "amenity",
        ...     "has",
        ...     functions.first,
        ...     with_forward_fill=True
        ...)
        >>> pivoted.orderBy("ts", "id").show()
        +---+---+-------+------+----+-----+
        | id| ts|balcony|fridge|oven| pool|
        +---+---+-------+------+----+-----+
        |  1|  1|   null| false|true|false|
        |  2|  2|  false|  null|null| null|
        |  1|  3|   null| false|true|false|
        |  1|  4|   null| false|true| true|
        |  1|  5|   true| false|true| true|
        +---+---+-------+------+----+-----+

        Great! Now every amenity that didn't have been changed kept it's state. BUT,
        the force change to null for amenity "oven" on id=1 at ts=4 was ignored during
        forward fill. If the user wants to respect this change, it must provide a mock
        value and type to be used as a signal for "true nulls". In other words, we want
        to forward fill only nulls that were created by the pivot transformation.

        In this example, amenities only assume boolean values. So there is no mock
        values for a boolean. It is only true or false. So users can give a mock value
        of another type (for which the column can be cast to). Check this out:

        >>> pivoted = pivot(
        ...     dataframe,
        ...     ["id", "ts"],
        ...     "amenity",
        ...     "has",
        ...     functions.first,
        ...     with_forward_fill=True,
        ...     mock_value=-1,
        ...     mock_type="int"
        ...)
        >>> pivoted.orderBy("ts", "id").show()
        +---+---+-------+------+----+-----+
        | id| ts|balcony|fridge|oven| pool|
        +---+---+-------+------+----+-----+
        |  1|  1|   null| false|true|false|
        |  2|  2|  false|  null|null| null|
        |  1|  3|   null| false|true|false|
        |  1|  4|   null| false|null| true|
        |  1|  5|   true| false|null| true|
        +---+---+-------+------+----+-----+

        During transformation, this method will cast the agg_column to mock_type
        data type and fill all "true nulls" with the mock_value. After pivot and forward
        fill are applied, all new pivoted columns will then return to the original type
        with all mock values replaced by null.
    """
    agg_column_type = None

    if mock_value is not None:
        if mock_type is None:
            raise AttributeError(
                "When proving a mock value, users must inform the data type,"
                " which should be supported by Spark."
            )
        agg_column_type = dict(dataframe.dtypes).get(agg_column)
        dataframe = dataframe.withColumn(
            agg_column, functions.col(agg_column).cast(mock_type)
        ).fillna({agg_column: mock_value})

    pivoted = (
        dataframe.groupBy(*group_by_columns)
        .pivot(pivot_column)
        .agg(aggregation(agg_column))
    )

    new_columns = [c for c in pivoted.columns if c not in group_by_columns]

    if with_forward_fill:
        for c in new_columns:
            pivoted = forward_fill(
                dataframe=pivoted,
                partition_by=group_by_columns[:-1],
                order_by=group_by_columns[-1],
                fill_column=c,
            )

    if mock_value is not None:
        for c in new_columns:
            pivoted = pivoted.withColumn(
                c,
                functions.when(functions.col(c) != mock_value, functions.col(c)).cast(
                    agg_column_type  # type: ignore
                ),
            )
    return pivoted



"""Holds the Sink class."""
from typing import List, Optional

from pyspark.sql.dataframe import DataFrame
from pyspark.sql.streaming import StreamingQuery

from butterfree.clients import SparkClient
from butterfree.hooks import HookableComponent
from butterfree.load.writers.writer import Writer
from butterfree.transform import FeatureSet
from butterfree.validations import BasicValidation
from butterfree.validations.validation import Validation


class Sink(HookableComponent):
    """Define the destinations for the feature set pipeline.

    A Sink is created from a set of writers. The main goal of the Sink is to
    trigger the load in each defined writers. After the load the entity can be
    used to make sure that all data was written properly using the validate
    method.

    Attributes:
        writers: list of Writers to use to load the data.
        validation: validation to check the data before starting to write.

    """

    def __init__(self, writers: List[Writer], validation: Optional[Validation] = None):
        super().__init__()
        self.enable_post_hooks = False
        self.writers = writers
        self.validation = validation

    @property
    def writers(self) -> List[Writer]:
        """List of Writers to use to load the data."""
        return self._writers

    @writers.setter
    def writers(self, value: List[Writer]) -> None:
        if not value or not all(isinstance(writer, Writer) for writer in value):
            raise ValueError("Writers needs to be a list of Writer instances.")
        else:
            self._writers = value

    @property
    def validation(self) -> Optional[Validation]:
        """Validation to check the data before starting to write."""
        return self._validation

    @validation.setter
    def validation(self, value: Validation) -> None:
        self._validation = value or BasicValidation()

    def validate(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient
    ) -> None:
        """Trigger a validation job in all the defined Writers.

        Args:
            dataframe: spark dataframe containing data from a feature set.
            feature_set: object processed with feature set metadata.
            spark_client: client used to run a query.

        Raises:
            RuntimeError: if any on the Writers returns a failed validation.

        """
        failures = []
        for writer in self.writers:
            try:
                writer.validate(
                    feature_set=feature_set,
                    dataframe=dataframe,
                    spark_client=spark_client,
                )
            except AssertionError as e:
                failures.append(e)

        if failures:
            raise RuntimeError(
                "The following validations returned error: {}".format(failures)
            )

    def flush(
        self, feature_set: FeatureSet, dataframe: DataFrame, spark_client: SparkClient
    ) -> List[StreamingQuery]:
        """Trigger a write job in all the defined Writers.

        Args:
            dataframe: spark dataframe containing data from a feature set.
            feature_set: object processed with feature set metadata.
            spark_client: client used to run a query.

        Returns:
            Streaming handlers for each defined writer, if writing streaming dfs.

        """
        pre_hook_df = self.run_pre_hooks(dataframe)

        if self.validation is not None:
            self.validation.input(pre_hook_df).check()

        handlers = [
            writer.write(
                feature_set=feature_set,
                dataframe=pre_hook_df,
                spark_client=spark_client,
            )
            for writer in self.writers
        ]

        return [handler for handler in handlers if handler]
